import asyncio
import inspect

from aiohttp import web as aio_web

from bolinette import env, core
from bolinette.commands import commands


class Bolinette:
    def __init__(self, *, profile=None, overrides=None):
        env.init_app(profile=profile, overrides=overrides)

        self.app = aio_web.Application()
        self.context = core.BolinetteContext(self.app)
        self.app['blnt'] = self.context

        self.run_init_functions(self.app)

    @staticmethod
    def run_init_functions(app):
        for func in core.cache.init_funcs:
            func(app['blnt'])

    def run(self):
        aio_web.run_app(self.app, port=self.context.env['port'])

    def run_command(self, name, **kwargs):
        kwargs['blnt'] = self
        if name in commands.commands:
            func = commands.commands[name]
            if inspect.isfunction(func):
                if inspect.iscoroutinefunction(func):
                    loop = asyncio.get_event_loop()
                    loop.run_until_complete(func(self.context, **kwargs))
                else:
                    func(**kwargs)



from bolinette import types


class MappingObject:
    def __init__(self, *, key=None, name=None, default=None, required=False, function=False, formatting=False):
        self.key = key
        self.name = name
        self.default = default
        self.required = required
        self.function = function
        self.formatting = formatting
        if not self.name and self.key:
            self.name = self.key


class Field(MappingObject):
    def __init__(self, field_type, *, key=None, name=None, default=None,
                 required=False, function=None, formatting=None):
        super().__init__(key=key, name=name, default=default, required=required,
                         function=function, formatting=formatting)
        self.type = field_type

    def __repr__(self):
        return f'<MappingField {self.key}:{self.type} -> {self.name}>'


class Column(Field):
    def __init__(self, column: 'types.defs.Column', *, name=None, default=None, required=False,
                 function=None, formatting=None):
        super().__init__(column.type, key=column.name, name=name, default=default, required=required,
                         function=function, formatting=formatting)


class List(MappingObject):
    def __init__(self, element, *, key=None, name=None, default=None,
                 required=False, function=None, formatting=None):
        super().__init__(key=key, name=name, default=default, required=required,
                         function=function, formatting=formatting)
        self.element = element

    def __repr__(self):
        return f'<MappingList {self.name}:[{repr(self.element)}]>'


class Definition(MappingObject):
    def __init__(self, model_name, model_key='default', *, key=None, name=None, default=None,
                 required=False, function=None, formatting=None):
        super().__init__(key=key, name=name, default=default, required=required,
                         function=function, formatting=formatting)
        self.fields = []
        self.model_name = model_name
        self.model_key = model_key

    def __repr__(self):
        return f'<MappingModel {self.model_name}.{self.model_key}>'


class Reference(Definition):
    def __init__(self, relationship: 'types.defs.Relationship', model_key='default', *, name=None, default=None,
                 required=False, function=None, formatting=None):
        super().__init__(relationship.model_name, model_key, key=relationship.name, name=name, default=default,
                         required=required, function=function, formatting=formatting)
        self.foreign_key = relationship.foreign_key.name
        self.reference_model = relationship.foreign_key.reference.model_name
        self.reference_key = relationship.foreign_key.reference.column_name



from typing import Dict

import aiohttp_cors
from aiohttp import web as aio_web
from aiohttp.web_request import Request
from aiohttp.web_urldispatcher import Resource, ResourceRoute

from bolinette import core, blnt, types
from bolinette.exceptions import APIError, ForbiddenError
from bolinette.utils import Pagination, response, functions, APIResponse
from bolinette.utils.serializing import deserialize, serialize


class BolinetteResources:
    def __init__(self, context: 'core.BolinetteContext'):
        self.context = context
        self._resources: Dict[str, 'BolinetteResource'] = {}
        self.cors = aiohttp_cors.setup(self.context.app, defaults={
            "*": aiohttp_cors.ResourceOptions(
                allow_credentials=True,
                expose_headers="*",
                allow_headers="*",
            )
        })

    def add_route(self, path: str, controller: 'blnt.Controller', route: 'blnt.ControllerRoute'):
        if path not in self._resources:
            self._resources[path] = BolinetteResource(self.cors.add(self.context.app.router.add_resource(path)))
        handler = RouteHandler(controller, route)
        self._resources[path].routes[route.method] = self.cors.add(
            self._resources[path].resource.add_route(route.method.http_verb, handler))


class BolinetteResource:
    def __init__(self, resource: Resource):
        self.resource = resource
        self.routes: Dict[types.web.HttpMethod, ResourceRoute] = {}


class RouteHandler:
    def __init__(self, controller: 'blnt.Controller', route: 'blnt.ControllerRoute'):
        self.controller = controller
        self.route = route

    async def __call__(self, request: Request):
        context: core.BolinetteContext = request.app['blnt']
        user_service = context.service('user')
        current_user = None
        payload = await deserialize(request)
        match = {}
        query = {}
        for key in request.match_info:
            match[key] = request.match_info[key]
        for key in request.query:
            query[key] = request.query[key]
        try:
            with core.Transaction(context):
                if self.route.access is not None:
                    current_user = await user_service.get_by_username(self.route.access.check(context, request),
                                                                      safe=True)

                if self.route.roles and len(self.route.roles):
                    user_roles = set(map(lambda r: r.name, current_user.roles))
                    if 'root' not in user_roles and not len(user_roles.intersection(set(self.route.roles))):
                        raise ForbiddenError(f'user.forbidden:{",".join(self.route.roles)}')

                if self.route.expects is not None:
                    exp_def = context.mapping.payload(self.route.expects.model, self.route.expects.key)
                    payload = context.mapping.validate_payload(exp_def, payload, self.route.expects.patch)
                    await context.mapping.link_foreign_entities(exp_def, payload)

                resp = await functions.async_invoke(self.route.func, self.controller, payload=payload,
                                                    match=match, query=query, current_user=current_user)

            if resp is None:
                return aio_web.Response(status=204)
            if isinstance(resp, aio_web.Response):
                return resp
            if isinstance(resp, str):
                return aio_web.Response(text=resp, status=200, content_type='text/plain')
            if not isinstance(resp, APIResponse):
                return aio_web.Response(text='global.unserializable_response', status=500, content_type='text/plain')

            content = resp.content

            if content.get('data') is not None and isinstance(content['data'], Pagination):
                content['pagination'] = {
                    'page': content['data'].page,
                    'per_page': content['data'].per_page,
                    'total': content['data'].total,
                }
                content['data'] = content['data'].items

            if self.route.returns is not None:
                ret_def = context.mapping.response(self.route.returns.model, self.route.returns.key)
                if content.get('data') is not None:
                    content['data'] = context.mapping.marshall(ret_def, content['data'],
                                                               skip_none=self.route.returns.skip_none,
                                                               as_list=self.route.returns.as_list)

            serialized, mime = serialize(content, 'application/json')

            web_response = aio_web.Response(text=serialized, status=resp.code, content_type=mime)
            for cookie in resp.cookies:
                if not cookie.delete:
                    web_response.set_cookie(cookie.name, cookie.value,
                                            expires=cookie.expires.strftime('%a, %d %b %Y %H:%M:%S GMT'),
                                            path=cookie.path, httponly=cookie.http_only)
                else:
                    web_response.del_cookie(cookie.name, path=cookie.path)

            return web_response
        except APIError as ex:
            res = response.from_exception(ex)
            serialized, mime = serialize(res.content, 'application/json')
            web_response = aio_web.Response(text=serialized, status=res.code, content_type=mime)
            return web_response



import requests
from bolinette_common import logger

from bolinette import env


class Mailgun:
    def __init__(self):
        self.url = env['MAILGUN_URL']
        self.ready = True
        self.key = env['MAILGUN_API']
        self.from_adr = env['MAILGUN_FROM']
        self._validate_attrs()

    def _validate_attrs(self):
        if not self.url:
            self.ready = False
            logger.error('MAILGUN_URL env key not set')
        if not self.key:
            self.ready = False
            logger.error('MAILGUN_API env key not set')
        if not self.from_adr:
            self.ready = False
            logger.error('MAILGUN_FROM env key not set')

    def send(self, to, subject, content):
        if self.ready:
            return requests.post(self.url, auth=("api", self.key), data={
                'from': self.from_adr,
                'to': list(to),
                'subject': subject,
                'text': content
            })



from typing import Dict, Any

from aiohttp import web as aio_web

from bolinette import core, env, blnt


class BolinetteContext:
    def __init__(self, app: aio_web.Application):
        self.app = app
        self.env = env
        self.db = core.DatabaseEngine(self)
        self.jwt = core.JWT()
        self.resources = core.BolinetteResources(self)
        self.sockets = core.BolinetteSockets(self)
        self.mapping = core.Mapping(self)
        self._tables: Dict[str, Any] = {}
        self._models: Dict[str, 'blnt.Model'] = {}
        self._repos: Dict[str, 'blnt.Repository'] = {}
        self._services: Dict[str, 'blnt.Service'] = {}
        self._controllers: Dict[str, 'blnt.Controller'] = {}
        self._ctx = {}

    def __getitem__(self, key):
        return self._ctx[key]

    def __setitem__(self, key, value):
        self._ctx[key] = value

    def model(self, name) -> Any:
        return self._models[name]

    def add_model(self, name, model: 'blnt.Model'):
        self._models[name] = model

    @property
    def models(self):
        return iter(self._models.items())

    def table(self, name) -> Any:
        return self._tables[name]

    def add_table(self, name, table):
        self._tables[name] = table

    @property
    def tables(self):
        return iter(self._tables.items())

    def repo(self, name) -> Any:
        return self._repos[name]

    def add_repo(self, name, repo: 'blnt.Repository'):
        self._repos[name] = repo

    @property
    def repos(self):
        return iter(self._repos.items())

    def service(self, name) -> Any:
        return self._services[name]

    def add_service(self, name, service: 'blnt.Service'):
        self._services[name] = service

    @property
    def services(self):
        return iter(self._services.items())

    def controller(self, name) -> Any:
        return self._controllers[name]

    def add_controller(self, name, controller: 'blnt.Controller'):
        self._controllers[name] = controller

    @property
    def controllers(self):
        return iter(self._repos.items())



import traceback

from bolinette_common import logger
from sqlalchemy.exc import SQLAlchemyError

from bolinette import core
from bolinette.exceptions import APIError, InternalError


class Transaction:
    def __init__(self, context: 'core.BolinetteContext'):
        self.context = context

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            self.context.db.session.rollback()
            if not issubclass(exc_type, APIError):
                logger.error(str(exc_val))
                traceback.print_tb(exc_tb)
                raise InternalError([str(exc_val)] + traceback.format_list(traceback.extract_tb(exc_tb)))
        else:
            try:
                self.context.db.session.commit()
            except SQLAlchemyError as err:
                self.context.db.session.rollback()
                raise InternalError([f'global.internal_error:{err}'])



from functools import wraps
from typing import Callable, Awaitable

import pytest

from bolinette import Bolinette, core
from bolinette.testing import TestClient, Mock


@pytest.fixture
def client(loop, aiohttp_client):
    blnt = Bolinette(profile='test',
                     overrides={'DBMS': 'SQLITE', 'SECRET_KEY': 'super secret'})
    client = loop.run_until_complete(aiohttp_client(blnt.app))
    return TestClient(client, blnt.context)


def bolitest(*, before: Callable[[core.BolinetteContext, Mock], Awaitable[None]] = None,
             after: Callable[[core.BolinetteContext, Mock], Awaitable[None]] = None):
    def wrapper(func):
        @wraps(func)
        async def inner(client: TestClient):
            await client.context.db.drop_all()
            await client.context.db.create_all()
            if before is not None:
                await before(client.context, client.mock)
            client.context.db.session.commit()
            await func(client=client)
            if after is not None:
                await after(client.context, client.mock)
            await client.context.db.drop_all()
        return inner
    return wrapper



from bolinette import blnt
from bolinette.decorators import controller, get
from bolinette.defaults.services import TimezoneService
from bolinette.utils import response


@controller('tz', '/tz')
class TimezoneController(blnt.Controller):
    @property
    def tz_service(self) -> TimezoneService:
        return self.context.service('tz')

    @get('')
    async def all_timezones(self):
        return response.ok('OK', await self.tz_service.get_all())



from bolinette import blnt
from bolinette.decorators import controller


@controller('role', '/role')
class RoleController(blnt.Controller):
    def default_routes(self):
        return [
            self.defaults.get_all(roles=['admin']),
            self.defaults.get_one('complete', key='name', roles=['admin'])
        ]



from bolinette.core.cache import cache
from bolinette.core.database import DatabaseEngine
from bolinette.core.jwt import JWT
from bolinette.core.resources import BolinetteResources, BolinetteResource
from bolinette.core.sockets import BolinetteSockets
from bolinette.core.mapping import Mapping
from bolinette.core.context import BolinetteContext
from bolinette.core.transaction import Transaction



from typing import List, Callable, Dict, Type

from bolinette import core, blnt


class BolinetteCache:
    def __init__(self):
        self.models: Dict[str, Type['blnt.Model']] = {}
        self.mixins: Dict[str, Type['blnt.Mixin']] = {}
        self.services: Dict[str, Type['blnt.Service']] = {}
        self.controllers: Dict[str, Type['blnt.Controller']] = {}
        self.topics: Dict[str, Type['blnt.Topic']] = {}
        self.init_funcs: List[Callable[['core.BolinetteContext'], None]] = []
        self.seeders = []


cache = BolinetteCache()



from bolinette import types, blnt
from bolinette.decorators import model


@model('file')
class File(blnt.Model):
    id = types.defs.Column(types.db.Integer, primary_key=True)
    key = types.defs.Column(types.db.String, nullable=False)
    name = types.defs.Column(types.db.String, nullable=False)
    mime = types.defs.Column(types.db.String, nullable=False)

    @classmethod
    def responses(cls):
        yield [
            types.mapping.Column(cls.key),
            types.mapping.Column(cls.name),
            types.mapping.Column(cls.mime)
        ]
        yield 'minimal', [
            types.mapping.Column(cls.key)
        ]



import os
from typing import Dict, Any

import yaml
from bolinette_common import paths


class Settings:
    def __init__(self):
        self._settings: Dict[str, Any] = {}

    def __getitem__(self, key: str):
        return self._settings.get(key.lower(), None)

    def __contains__(self, key: str):
        return key.lower() in self._settings

    def get(self, key: str, default=None):
        item = self[key]
        return item if item is not None else default

    def _reset(self, settings: Dict[str, Any]):
        self._settings = {}
        for key, value in settings.items():
            self._settings[key.lower()] = value


class Environment(Settings):
    def __init__(self):
        super().__init__()
        self.cwd = paths.cwd()
        self.origin = paths.dirname(__file__)
        self.init = Settings()
        self.init._reset(self.load_from_file('init.yaml'))

    def instance_path(self, *path):
        return self.root_path('instance', *path)

    def root_path(self, *path):
        return paths.join(self.cwd, *path)

    def internal_path(self, *path):
        return paths.join(self.origin, *path)

    @property
    def default_env(self):
        return {
            'APP_NAME': 'DEFAULT_NAME',
            'APP_DESC': 'DEFAULT_DESCRIPTION',
            'APP_VERSION': '0.0.1',
            'DBMS': 'SQLITE',
            'DEBUG': True,
            'HOST': '127.0.0.1',
            'PORT': '5000',
            'WEBAPP_FOLDER': self.root_path('webapp', 'dist')
        }

    def read_profile(self):
        try:
            with open(self.instance_path('.profile')) as f:
                for line in f:
                    return line.strip().replace('\n', '')
        except FileNotFoundError:
            return None

    def init_app(self, *, profile=None, overrides=None):
        profile = profile or self.read_profile() or 'development'
        self._reset(self.merge_env_stack([
            self.default_env,
            self.load_from_file(f'env.{profile}.yaml'),
            self.load_from_file(f'env.local.{profile}.yaml'),
            self.load_from_os(),
            overrides or {},
            {'profile': profile}
        ]))

    def load_from_os(self):
        keys = {}
        for key in os.environ:
            if key.startswith('BLNT_'):
                keys[key[5:]] = os.environ[key]
        return keys

    def load_from_file(self, file_name):
        try:
            with open(self.instance_path(file_name), 'r') as f:
                return yaml.safe_load(f)
        except FileNotFoundError:
            return {}

    def merge_env_stack(self, stack):
        settings = {}
        for source in stack:
            for key, value in source.items():
                settings[key.lower()] = value
        return settings


env = Environment()



import json

from bolinette import core
from bolinette.testing import Mock
from bolinette.utils.serializing import serialize


class TestClient:
    def __init__(self, client, context: core.BolinetteContext):
        self.client = client
        self.context = context
        self.mock = Mock(context)
        self.cookies = {}

    def payload(self, **payload):
        if 'csrf_access_token' in self.cookies:
            payload['headers'] = {}
            payload['headers']['X-CSRF-TOKEN'] = self.cookies['csrf_access_token']
        return payload

    def parse_cookies(self, headers):
        for key, header in headers.items():
            if key == 'Set-Cookie':
                split = header.split(';')[0].split('=')
                self.cookies[split[0]] = split[1]

    async def post(self, path: str, data: dict = None) -> dict:
        if data is None:
            data = {}
        res = await self.client.post(f'/api{path}', data=serialize(data, 'application/json')[0],
                                     headers={'Content-Type': 'application/json'})
        self.parse_cookies(res.headers)
        return json.loads(await res.text())

    async def put(self, path: str, data: dict = None) -> dict:
        if data is None:
            data = {}
        res = await self.client.put(f'/api{path}', data=serialize(data, 'application/json')[0],
                                    headers={'Content-Type': 'application/json'})
        self.parse_cookies(res.headers)
        return json.loads(await res.text())

    async def patch(self, path: str, data: dict = None) -> dict:
        if data is None:
            data = {}
        res = await self.client.patch(f'/api{path}', data=serialize(data, 'application/json')[0],
                                      headers={'Content-Type': 'application/json'})
        self.parse_cookies(res.headers)
        return json.loads(await res.text())

    async def get(self, path: str) -> dict:
        res = await self.client.get(f'/api{path}')
        return json.loads(await res.text())

    async def delete(self, path: str) -> dict:
        res = await self.client.delete(f'/api{path}')
        return json.loads(await res.text())



from bolinette_common import logger
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

from bolinette import core


class DatabaseEngine:
    def __init__(self, context: 'core.BolinetteContext'):
        self.engine = create_engine(self._create_uri(context), echo=False)
        self.Session = sessionmaker()
        self.model = declarative_base()
        self.Session.configure(bind=self.engine)
        self.session = self.Session()

    async def create_all(self):
        self.model.metadata.create_all(self.engine)

    async def drop_all(self):
        self.model.metadata.drop_all(self.engine)

    def _create_uri(self, context: 'core.BolinetteContext'):
        dbms = context.env.get('DBMS', 'SQLITE').lower()
        if dbms == 'sqlite':
            return 'sqlite:///' + context.env.instance_path(
                context.env.get('SQLITE_FILE', f'{context.env["PROFILE"]}.db'))
        if dbms == 'memory':
            return 'sqlite://'
        if dbms == 'postgresql':
            return 'postgresql://' + context.env['DB_URL']
        logger.error(f'Unknown database system "{dbms}"')
        exit(1)

    async def run_seeders(self, context: 'core.BolinetteContext'):
        for func in core.cache.seeders:
            await func(context)



from typing import Callable, List, Dict

from bolinette import core, types, blnt
from bolinette.utils import response, functions


class Controller:
    __blnt__: 'ControllerMetadata' = None

    def __init__(self, context: 'core.BolinetteContext'):
        self.__props__ = ControllerProps(self)
        self.context = context
        self.defaults = ControllerDefaults(self)

    @property
    def service(self) -> blnt.Service:
        return self.context.service(self.__blnt__.service_name)

    def default_routes(self):
        return []

    def __repr__(self):
        return f'<Controller {self.__blnt__.name} {self.__blnt__.path}>'


class SimpleController:
    __blnt__: 'ControllerMetadata' = None

    def __init__(self, context: 'core.BolinetteContext'):
        self.context = context
        self.__props__ = ControllerProps(self)

    def __repr__(self):
        return f'<Controller {self.__blnt__.name} {self.__blnt__.path}>'


class ControllerMetadata:
    def __init__(self, name: str, path: str, service_name: str):
        self.name = name
        self.path = path
        self.service_name = service_name


class ControllerProps:
    def __init__(self, controller):
        self.controller = controller

    def _get_attribute_of_type(self, attr_type):
        return dict([(name, attribute)
                     for name, attribute in vars(self.controller.__class__).items()
                     if isinstance(attribute, attr_type)])

    def get_routes(self) -> Dict[str, 'ControllerRoute']:
        return self._get_attribute_of_type(ControllerRoute)


class ControllerRoute:
    def __init__(self, func: Callable, path: str, method: types.web.HttpMethod,
                 access: types.web.AccessToken = None, expects: 'ControllerExcepts' = None,
                 returns: 'ControllerReturns' = None, roles: List[str] = None):
        self.controller = None
        self.func = func
        self.path = path
        self.method = method
        self.access = access or types.web.AccessToken.Optional
        self.expects = expects
        self.returns = returns
        self.roles = roles or []


class ControllerExcepts:
    def __init__(self, model: str, key: str = 'default', *, patch: bool = False):
        self.model = model
        self.key = key
        self.patch = patch


class ControllerReturns:
    def __init__(self, model: str, key: str = 'default', *, as_list: bool = False, skip_none: bool = False):
        self.model = model
        self.key = key
        self.as_list = as_list
        self.skip_none = skip_none


class ControllerDefaults:
    def __init__(self, controller: Controller):
        self.service: blnt.Service = controller.service

    def get_all(self, returns='default', *, access=None, roles=None):
        async def route(controller, *, query, **kwargs):
            pagination = None
            order_by = []
            if 'page' in query or 'per_page' in query:
                pagination = {
                    'page': int(query.get('page', 0)),
                    'per_page': int(query.get('per_page', 20))
                }
            if 'order_by' in query:
                columns = query['order_by'].split(',')
                for column in columns:
                    order_args = column.split(':')
                    col_name = order_args[0]
                    order_way = order_args[1] if len(order_args) > 1 else 'asc'
                    order_by.append((col_name, order_way == 'asc'))
            resp = await functions.async_invoke(controller.service.get_all,
                                                pagination=pagination, roles=roles, **kwargs)
            return response.ok('OK', resp)

        return ControllerRoute(route, '', types.web.HttpMethod.GET, access=access, roles=roles,
                               returns=ControllerReturns(self.service.__blnt__.model_name, returns, as_list=True))

    def get_one(self, returns='default', *, key='id', access=None, roles=None):
        async def route(controller, *, match, **kwargs):
            resp = await functions.async_invoke(controller.service.get_first_by, key, match.get('value'), **kwargs)
            return response.ok('OK', resp)

        return ControllerRoute(route, '/{value}', types.web.HttpMethod.GET, access=access, roles=roles,
                               returns=ControllerReturns(self.service.__blnt__.model_name, returns))

    def create(self, returns='default', expects='default', *, access=None, roles=None):
        async def route(controller, payload, **kwargs):
            resp = await functions.async_invoke(controller.service.create, payload, **kwargs)
            return response.created(f'{controller.service.__blnt__.model_name}.created', resp)

        return ControllerRoute(route, '', types.web.HttpMethod.POST, access=access, roles=roles,
                               expects=ControllerExcepts(self.service.__blnt__.model_name, expects),
                               returns=ControllerReturns(self.service.__blnt__.model_name, returns))

    def update(self, returns='default', expects='default', *, key='id', access=None, roles=None):
        async def route(controller, payload, match, **kwargs):
            entity = await controller.service.get_first_by(key, match.get('value'))
            resp = await functions.async_invoke(controller.service.update, entity, payload, **kwargs)
            return response.ok(f'{controller.service.__blnt__.model_name}.updated', resp)

        return ControllerRoute(route, '/{value}', types.web.HttpMethod.PUT, access=access, roles=roles,
                               expects=ControllerExcepts(self.service.__blnt__.model_name, expects),
                               returns=ControllerReturns(self.service.__blnt__.model_name, returns))

    def patch(self, returns='default', expects='default', *, key='id', access=None, roles=None):
        async def route(controller, payload, match, **kwargs):
            entity = await controller.service.get_first_by(key, match.get('value'))
            resp = await functions.async_invoke(controller.service.patch, entity, payload, **kwargs)
            return response.ok(f'{controller.service.__blnt__.model_name}.updated', resp)

        return ControllerRoute(route, '/{value}', types.web.HttpMethod.PATCH, access=access, roles=roles,
                               expects=ControllerExcepts(self.service.__blnt__.model_name, expects, patch=True),
                               returns=ControllerReturns(self.service.__blnt__.model_name, returns))

    def delete(self, returns='default', *, key='id', access=None, roles=None):
        async def route(controller, match, **kwargs):
            entity = await controller.service.get_first_by(key, match.get('value'))
            resp = await functions.async_invoke(controller.service.delete, entity, **kwargs)
            return response.ok(f'{controller.service.__blnt__.model_name}.deleted', resp)

        return ControllerRoute(route, '/{value}', types.web.HttpMethod.DELETE, access=access, roles=roles,
                               returns=ControllerReturns(self.service.__blnt__.model_name, returns))



import bcrypt as py_bcrypt

from bolinette import env


class Bcrypt:
    def __init__(self):
        self.secret_key = None

    def init_app(self):
        self.secret_key = env['SECRET_KEY']

    def hash_password(self, password):
        return py_bcrypt.hashpw(password.encode(), py_bcrypt.gensalt())

    def check_password(self, hashed, password):
        return py_bcrypt.checkpw(password.encode(), hashed)


bcrypt = Bcrypt()
